""" Intent matching functions that return true or false for the intent matching the statement

(dependency injections for intent recognitizer check_match)
"""
from .classifiers import classify_intent


def exact(statement, intent):
    return statement == intent


def fuzzy(statement, intent):
    """ Use kalika's emotional word/intent recognizer """
    return classify_intent(statement) == intent


def keyword(statement, intent):
    """ Use kalika's emotional word/intent recognizer """
    # TODO: tokenize statement with memoized tokenizer
    return intent in statement


def default(statement, intent):
    return exact(statement, intent)





""" Base classes for chatbot skills answering questions and executing rule-based dialog """
from copy import copy
import logging
import os
import uuid
from collections import namedtuple, abc
import zipfile

import pandas as pd
# from multiprocessing import cpu_count

from qary.etl.netutils import download_if_necessary
from qary.config import MIDATA_URL, MIDATA_QA_MODEL_DIR, DEFAULT_SKILL_CONFIDENCE, CLI_ARGS  # , USE_CUDA
from qary.etl.fileutils import LARGE_FILES
from qary.etl.nesting import dict_merge


log = logging.getLogger(__name__)

# tuple to hold confidence, text, and bot-name
BotReply_fields = ('confidence', 'text', 'skill', 'context')
BotReply = namedtuple('BotReply', BotReply_fields)  # noqa
BotReply.__new__.__defaults__ = (None,) * len(BotReply_fields)


def normalize_replies(replies):
    """ Make sure a list of replies includes score and text. TODO: confidences sum to 1

    >>> normalize_replies(['hello world']) == [BotReply(confidence=DEFAULT_SKILL_CONFIDENCE, text='hello world')]
    True
    >>> normalize_replies([])
    []
    >>> normalize_replies([BotReply()]) == [BotReply()]
    True
    >>> normalize_replies([(1, 'one')]) == [BotReply(1, 'one')]
    True
    >>> normalize_replies([(1, 'one'), BotReply(2.0, 'two')]) == [BotReply(1, 'one'), BotReply(2.0, 'two')]
    True
    """
    if isinstance(replies, BotReply):
        replies = [replies]
    elif isinstance(replies, str):
        replies = [(DEFAULT_SKILL_CONFIDENCE, replies)]
    elif isinstance(replies, (tuple, list)):
        if len(replies) == 1 and isinstance(replies[0], str):
            replies = [(DEFAULT_SKILL_CONFIDENCE, replies[0])]
        elif len(replies) == 2:
            if (isinstance(replies[0], (float, int)) and isinstance(replies[1], str)):
                replies = [BotReply(replies[0], replies[1])]
            elif (isinstance(replies[1], (float, int)) and isinstance(replies[0], str)):
                replies = [BotReply(replies[1], replies[0])]
    else:
        raise RuntimeError(
            f"The conversation manager requires a list or tuple of possible replies. Received {type(replies)}")
    # TODO: this sorting is likely unnecessary, redundant with sort happening within CLIBot.reply()
    normalized_replies = []
    for r in replies:
        if isinstance(r, BotReply):
            normalized_replies.append(r)
        elif isinstance(r, dict):
            normalized_replies.append(BotReply(**r))
        else:
            normalized_replies.append(BotReply(*r))
    return normalized_replies


class Replies(list):
    """ WIP: Container for 2-tuples of scored replies (strings)

    TODO:
        - override __getitem__ to return None when empty
        - validate method to make sure 2-tuples of float, str
        - normalize method to ensures scores add up to 1
        - softmax to improve likelihood of best replies
        - sort method?
    """
    pass


class ContextDict(dict):
    """ WIP: Container for context dict attributes for bots

    ContextDict.merge should use dict_merge

    SEE: https://stackoverflow.com/a/3233356/623735
    """

    def merge(self, other):
        dict_merge(self, other)

    def dict_merge(self, other):
        dict_merge(self, other)


class ContextSeries(pd.Series):
    """ WIP: Container for nested json-like object used to keep track of bot context

    API should be identical to ContextDict for .__getitem__
    dict-like object that iterates through the values of a mapping (like a Series)
    rather than the keys.
    """

    def __init__(self, obj, *args, dtype='O', **kwargs):
        super().__init__(obj, *args, dtype=dtype, **kwargs)

    def __iter__(self):
        for k, v in self.items():
            yield v


def update_dict(d, u):
    """ SEE: https://stackoverflow.com/a/3233356/623735 """
    for k, v in u.items():
        if isinstance(v, abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class EmptyRepliesBot:
    # def __init__(self, **kwargs):
    #     super().__init__()

    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> bot = EmptyRepliesBot()
        >>> bot.reply('Hi')
        []
        """
        return []


class HiBot(EmptyRepliesBot):
    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> bot = HiBot()
        >>> bot.reply('Hi')
        [(1.0, 'Hi!')]
        """
        s = statement.lower().strip().split()
        if s and s[0] in ('hi', 'hello', 'howdy', 'helo', 'yo'):
            return [(1.0, 'Hi!')]
        return []


class ContextBaseSkill:
    """ Manages self.context attribute with update_, reset_ and reply(context=...) methods

    >>> default_context = {'cli_args': {'config': None, 'debug': True, 'nickname': 'qary',
    ...     'num_top_replies': 10, 'use_cuda': False, 'persist': True, 'bots': ['glossary', 'faq', 'eliza'],
    ...     'verbosity': None, 'loglevel': 30, 'spacy_lang': 'en_core_web_md', 'wiki_title_max_words': 3,
    ...     'semantic': 1.0, 'spell': 0.2, 'words': [], 'qa_model': 'albert-large-v2-0.2.0',
    ...     'domains': None, 'glossary_domains': None, 'quiz_domains': None}}
    >>> conbot = ContextBaseSkill()
    >>> conbot.context == default_context
    True
    >>> default_context.update({'init': 'init_value'})
    >>> conbot = ContextBaseSkill({'init': 'init_value'})
    >>> conbot.context == default_context
    True
    >>> conbot.update_context({'new': 'new_value', 'cli_args': None, 'args': None, 'kwargs': None})
    {'cli_args': None, 'init': 'init_value', 'new': 'new_value', 'args': None, 'kwargs': None}
    >>> conbot.context
    {'cli_args': None, 'init': 'init_value', 'new': 'new_value', 'args': None, 'kwargs': None}
    >>> conbot.update_context({'init': 'updated_existing', 'new': {'inner': 'new_innards'}})
    {'cli_args': None, 'init': 'updated_existing', 'new': {'inner': 'new_innards'}, 'args': None, 'kwargs': None}
    >>> conbot.context
    {'cli_args': None, 'init': 'updated_existing', 'new': {'inner': 'new_innards'}, 'args': None, 'kwargs': None}
    >>> conbot.reply('Hi', context={'new': {'inner_reply': 'new_inner_reply'}})
    []
    >>> conbot.context['init']
    'updated_existing'
    >>> conbot.context['new']
    {'inner': 'new_innards', 'inner_reply': 'new_inner_reply'}
    """

    def __init__(self, context=None, cli_args=CLI_ARGS, *args, **kwargs):
        # FIXME: ContextBaseSkill should only handle context, nothing else
        self.context = {}
        self.update_context(kwargs or {})
        try:
            cli_args = vars(cli_args)
        except AttributeError:
            cli_args = dict(cli_args or {})
        self.update_context(context={'cli_args': cli_args})
        if isinstance(context, str):
            log.warning("Deprecated API: `context` should be a nested dictionary. Texts belong at `context['doc']['text']`).")
            self.update_context({'doc': {'text': str(context)}})
        elif context and len(context):
            self.update_context(dict(context))
        super().__init__()

    def update_context(self, context=None):
        # defensive programming in case of: skill.context = None | ''
        self.context = getattr(self, 'context', {}) or {}
        if not isinstance(self.context, (abc.Mapping, ContextDict)):
            log.error(f'self.context is not a dict! self.context: {self.context}')
            self.context = dict(self.context)

        log.warning(f"Updating self.context using context: {repr(context)[:61]}...")
        if isinstance(context, str):
            context = {'doc': {'text': context}}
        context = {} if not context else dict(context)

        dict_merge(self.context, context)
        log.debug(f"Updated self.context: {repr(self.context)[:60]}")
        return self.context

    def reset_context(self, context=None):
        self.context = {k: copy(self.context[k]) for k in self.context if k in ['cli_args', 'args', 'kwargs']}
        log.info(f"Reset self.context to self.context={repr(self.context)[:60]} before updating with {repr(context)[:60]}")
        self.update_context(context=context)
        return self.context

    def reply(self, statement, context=None):
        """ Chatbot "main" function to respond to a user command or statement

        >>> s = ContextBaseSkill()
        >>> s.reply('Hi', context={'new': 'context'})
        []
        >>> s.context['new']
        'context'
        >>> list(s.context.keys())
        ['cli_args', 'new']
        """
        log.warning(f'base.ContextBaseSkill context={context}')
        self.update_context(context=context)
        return []


class HistoryBot:
    """ Remembers the history of every user statement and bot response string

    >>> histbot = HistoryBot()
    >>> histbot.history
    []
    >>> histbot.update_history('Hi')
    1
    >>> histbot.history
    [{'statement': 'Hi', 'possible_statements': (), 'agent': None}]
    >>> histbot.update_history('Hello', agent='human')
    2
    >>> histbot.history
    [{'statement': 'Hi', 'possible_statements': (), 'agent': None},
     {'statement': 'Hello', 'possible_statements': (), 'agent': 'human'}]
    """

    def __init__(self, history=None, **kwargs):
        super().__init__(**kwargs)
        self.history = history or []

    def update_history(self, statement=None, possible_statements=None, agent=None):
        """ Append convo history with a statement/reply, the possibilities it was chosen from, and an agent (bot/user) name """
        self.history.append(dict(
            statement=statement,
            possible_statements=tuple(possible_statements or []),
            agent=agent))
        return len(self.history)

    def reset_history(self, statement=None, possible_statements=None, agent=None):
        """ Reset conversation history to a single statement/reply

        Resets the conversation history to the last interaction:
            - a single most recent reply
            - the possibile Skill replies list that the reply was chosen from
            - the Skills that created the list of possible replies
        """
        self.history = []
        if statement or possible_statements or agent:
            self.update_history(
                statement=statement,
                possible_statements=possible_statements,
                agent=agent)
        return len(self.history)


class TransformerBot(HistoryBot, ContextBaseSkill):
    """ Base Skill class that maintains context and load transformer models.  """

    def __init__(self, context=None, cli_args=CLI_ARGS, history=None, **kwargs):
        super().__init__(context=context, cli_args=CLI_ARGS, history=history, **kwargs)

    def load_model(self, args):
        self.transformer_loggers = []
        for name in log.root.manager.loggerDict:
            if (len(name) >= 12 and name[:12] == 'transformers') or name == 'qary.skills.qa.utils':
                self.transformer_loggers.append(log.getLogger(name))
                self.transformer_loggers[-1].setLevel(log.ERROR)

        qa_model = args.qa_model
        url_str = f"{MIDATA_URL}/{MIDATA_QA_MODEL_DIR}/{qa_model}.zip"
        log.warning(f"Attempting to download url: {url_str}")
        model_dir = os.path.dirname(LARGE_FILES['albert-large-v2']['path'])
        model_type = qa_model.split('-')[0].lower()
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

        if not all((
            os.path.exists(os.path.join(model_dir, 'config.json')),
            os.path.exists(os.path.join(model_dir, 'pytorch_model.bin')),
            os.path.exists(os.path.join(model_dir, 'tokenizer_config.json')),
            os.path.exists(os.path.join(model_dir, 'version.json')),
            any((model_type == 'bert' and os.path.exists(os.path.join(model_dir, 'vocab.txt'))),
                (model_type == 'albert' and os.path.exists(os.path.join(model_dir, 'spiece.model')))),
        )):
            zip_local_path = download_if_necessary(
                url=LARGE_FILES['albert-large-v2']['url'],
                path=LARGE_FILES['albert-large-v2']['path'])
            with zipfile.ZipFile(zip_local_path, 'r') as zip_file:
                zip_file.extractall(os.path.dirname(LARGE_FILES['albert-large-v2']['path']))
            os.remove(zip_local_path)

    def encode_transformer_input(self, statement, context=None):
        """ Convert statement and context strings into nested dict format compatible with [AL]BERT transformer

        >>> bot = TransformerBot()
        >>> encoded = bot.encode_transformer_input('statement', 'context')
        >>> encoded[0]['qas'][0]['question']
        'statement'
        >>> encoded[0]['context']
        'context'
        """
        if context is None:
            context = self.context
        if isinstance(context, str):
            # TODO try/except on [] instead of get to provide deprecation warnings
            text = context
            super().update_context({'doc': {'text': text}})
        else:
            super().update_context(context)
        text = self.context['doc']['text']
        encoded = [{
            'qas': [{
                'id': str(uuid.uuid1()),
                'question': statement
            }],
            'context': text}]
        return encoded

    def decode_transformer_output(self, output):
        """ Extract reply string from the transformer model's prediction output (nested dict)

        >>> bot = TransformerBot()
        >>> bot.decode_transformer_output([{'id': 'unique_id', 'answer': 'response', 'probability': 0.75}])
        (0.75, 'response')
        """
        return output[0]['probability'], output[0]['answer']



# -*- coding: utf-8 -*-
__version__ = 'unknown'

try:
    from pkg_resources import get_distribution, DistributionNotFound
except ImportError as e:
    print(e)

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound as e:
    print(e)
finally:
    del get_distribution, DistributionNotFound


# from qary.init import maybe_download  # noqa



""" notes

adventure = Adventure()


#  current state = [list, of, actions]
start = [adventure.cottage]
cottage = [adventure.hometown, adventure.back_to_full_health, adventure.check_stats]
hometown =  [adventure.cottage, adventure.forest, adventure.coast, adventure.check_stats]
forest =  [adventure.hometown, adventure.find_stones, adventure.battle, adventure.check_stats, adventure.forest]
battle = [adventure.forest, adventure.battle_injured, adventure.battle_injured]
coast = [adventure.hometown, adventure.fishconvo, adventure.find_stones, adventure.check_stats, adventure.coast]
fishconvo = [adventure.fishconvo_yes, adventure.fishconvo_no, adventure.coast, adventure.coast]


current_state_actions = {"start": [adventure.cottage],
                         "cottage": [adventure.hometown, adventure.back_to_full_health, adventure.check_stats],
                         "hometown": [adventure.cottage, adventure.forest, adventure.coast, adventure.check_stats],
                         "forest": [adventure.hometown, adventure.find_stones, adventure.battle, adventure.check_stats, adventure.forest],
                         "battle": [adventure.forest, adventure.battle_injured, adventure.battle_injured],
                         "coast": [adventure.hometown, adventure.fishconvo, adventure.find_stones, adventure.check_stats, adventure.coast],
                         "fishconvo": [adventure.fishconvo_yes, adventure.fishconvo_no, adventure.coast, adventure.coast]}


possible_actions = namedtuple("possible_actions", ["current_state", "condition", "action"])

data_table = [
        possible_actions(df['current_state'][0],
                   df['condition'][0] or None,
                   # getattr(adventure, df['action'][0], df['action'][0])),

        possible_actions(df['current_state'][1], df['condition'][1] or None, df['action'][0]),





# nested_dict = { 'dictA': {'key_1': 'value_1'},
#                 'dictB': {'key_2': 'value_2'}}


# current_state_actions = {"start": {'1': adventure.cottage},
#                          "cottage": {'1': adventure.hometown, '2': adventure.back_to_full_health, '9': adventure.check_stats,
#                          "hometown": [adventure.cottage, adventure.forest, adventure.coast, adventure.check_stats],
#                          "forest": [adventure.hometown, adventure.find_stones, adventure.battle, adventure.check_stats, adventure.forest],
#                          "battle": [adventure.forest, adventure.battle_injured, adventure.battle_injured],
#                          "coast": [adventure.hometown, adventure.fishconvo, adventure.find_stones, adventure.check_stats, adventure.coast],
#                          "fishconvo": [adventure.fishconvo_yes, adventure.fishconvo_no, adventure.coast, adventure.coast]}
"""



# qary.chat.v3
""" Load *.v3.dialog.yml files and execute recognize_intent() function:

  Inputs:
    lang (str): e.g. 'en', 'zh', 'es'
    state name (str): e.g. 'language-selected-english'
    user utterance (str): 'Hello chatobt'
  Outputs:
    id (str): state name
    text: bot utterance
"""
from collections import abc
import io
import logging
from pathlib import Path
from qary.init import maybe_download
import time
import yaml

import numpy as np
import pandas as pd

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

WELCOME_STATE_NAME = '__WELCOME__'
DEFAULT_LANG = 'en'
EXAMPLE_DIALOG_TREE_FILEPATH = maybe_download(filename=Path('chat/example-multilingual.v3.dialog.yml'))
EMPTY_DIALOG_TREE = pd.Series({
    WELCOME_STATE_NAME:
        {'name': WELCOME_STATE_NAME,
         'actions':
            {'en': "I have nothing to say ;-) (My v3 dialog tree is empty)"},
         'triggers':
            {'en': {"": WELCOME_STATE_NAME}},
         }
})

# TODO see ISO_LANG codes in constants for nessvec or qary package
LANGS = {
    'en': 'en',
    'english': 'en',
    'es': 'es',
    'spanish': 'es',
    'zh': 'zh',
    'zho': 'zh',
    'chinese': 'zh',
    'chinese-simplified': 'zh',
    'zht': 'zht',
    'chinese-traditional': 'zht',
    'traditional chinese': 'zht',
    'traditional-chinese': 'zht',
}


def normalize_state_name(name):
    return str(name).strip().lower()


def normalize_user_text(text):
    return str(text).strip().lower()


def normalize_intent_texts(states):
    for name, state in states.items():
        for lang, intents in state.get('triggers', {}).items():
            if lang not in LANGS:
                continue
            try:
                states[name]['triggers'][lang].update({
                    normalize_user_text(i): v for (i, v) in intents.items()
                })
            except AttributeError as e:
                log.warning(e)
                log.warning(f"states[name]['triggers'][lang] = {states[name]['triggers'][lang]}")
    return states


def update_node_with_buttons(node, **context):
    """ Use lang in context dict to add button text to bot_text attribute of node """
    lang = context.get('lang', DEFAULT_LANG)
    log.debug(node)
    lang_buttons = node.get('buttons', {})
    node['actions_with_buttons'] = {}
    for lang, buttons in lang_buttons.items():
        if isinstance(buttons, str):
            buttons = [buttons]
        buttons = list(buttons)

        log.debug('=' * 80)
        log.debug(f'Updating bot_text for "{lang}" in node named "{node["name"]}" with {len(buttons)} buttons')
        bot_text = node['actions'].get(lang, '')
        for button_text in buttons:
            bot_text += f'\n  > {button_text}'
        log.debug(bot_text)
        log.debug('=' * 80)
        node['actions_with_buttons'][lang] = bot_text
    log.debug(node)
    return node


def update_state_list_with_buttons(states):
    log.warning(type(states))
    if isinstance(states, abc.Mapping):
        for state_name in states:
            states[state_name] = update_node_with_buttons(states[state_name])
    elif isinstance(states, (list, pd.Series)):
        for i, node in enumerate(states):
            states[i] = update_node_with_buttons(node)

    return states


def coerce_dialog_tree_series(states: (str, Path, abc.Mapping, io.TextIOWrapper, pd.Series, list, np.ndarray, pd.Series)):
    """ Ensure that (yamlfilepath | list) -> dict -> Series """
    if isinstance(states, pd.Series):
        return states
    if isinstance(states, (str, Path)):
        with Path(states).open() as fin:
            return coerce_dialog_tree_series(yaml.full_load(fin))
    if isinstance(states, io.TextIOWrapper):  # opened file mode='rt'
        return coerce_dialog_tree_series(yaml.full_load(states))
    if isinstance(states, abc.Mapping):
        return pd.Series(states)
    if isinstance(states, (list, np.ndarray)):
        states = pd.Series(
            states,
            index=[s.get('name', str(i)) for (i, s) in enumerate(states)])
        states.index.name = 'name'
        return states
    if states is None:
        return EMPTY_DIALOG_TREE
    raise ValueError(
        f"Unable to coerce {type(states)} into pd.Series:\n  states={str(states)[:130]}..."
    )


EXAMPLE_DIALOG_TREE = coerce_dialog_tree_series(states=EXAMPLE_DIALOG_TREE_FILEPATH)
EXAMPLE_DIALOG_TREE_DICT = EXAMPLE_DIALOG_TREE.to_dict()
EXAMPLE_DIALOG_TREE_SERIES = EXAMPLE_DIALOG_TREE


def preprocess_dialog_tree_series(states=EXAMPLE_DIALOG_TREE):
    """ Ensure that all dialog tree index values are normalized strs: Series.index.str.lower().str.strip()."""

    states = coerce_dialog_tree_series(states)
    states = update_state_list_with_buttons(states)
    states = normalize_intent_texts(states)

    # make sure state['name'] is used to overwrite the index
    states.index = [normalize_state_name(s.get('name', i)) for (i, s) in states.items()]
    states.index.name = 'name'

    is_dupe = states.index.duplicated()
    if sum(is_dupe) > 0:
        log.error(f'State pd.Series contains {sum(is_dupe)} duplicate index values!\n'
                  f'The following duplicates will be deleted:\n{states.index[is_dupe].to_series()}'
                  )
    states = pd.Series(states[~is_dupe])
    states_list = []
    for i, (idx, s) in enumerate(states.items()):
        s = dict(s)
        s['name'] = str(idx)
        s['iloc'] = i
        states_list.append((idx, s))
        # states_list.append(i, s)
    return pd.Series(dict(states_list))


def await_timeout(seconds=0, milliseconds=0):
    time.sleep(seconds + milliseconds / 1000.0)


def next_state_context(states=EXAMPLE_DIALOG_TREE, state_name=None, user_text='', **context):
    r""" Recognize desired state transition trigger and return next_state name

      Inputs:
        lang (str): e.g. 'en', 'zh', 'es'
        state name (str): e.g. 'language-selected-english'
        user utterance (str): 'Hello chatobt'
      Outputs:
        name (str): state name
        text: bot utterance

    >>> next_state_context(state_name='is-this-your-first-time', user_text='yes')
    {'lang': 'en',
     'state_name': 'is-this-your-first-yes',
     'bot_text': 'Which of the following do you represent?\n'}
    """

    context['lang'] = context.get('lang', DEFAULT_LANG)

    state_name = states[0]['name'] if state_name is None else state_name
    # default_response = dict(
    #     state_name=states[0]['name'],
    #     bot_text=states[0].get('actions', {}).get(lang, states[0].get('en', '')),
    #     lang=lang)
    if state_name is None:
        return None
    # if str(state_name).lower().strip() in ('__start__', '__welcome__'):
    #     return default_response
    log.debug('\n' + str(states))

    # detect intent from text message
    try:
        node = states[state_name]
    except KeyError as e:
        log.error(e)
        log.error(f"Invalid state_name '{state_name}'. It is not among states.index:\n  {states.index}")

    if state_name != node['name']:
        log.debug(f' states[name]:{node.get("name")} != {state_name}:state_name')

    triggers = node.get('triggers', {})
    log.debug(f'    triggers for node name {node["name"]}:\n{triggers}')

    user_intents = triggers.get(context['lang'], triggers.get(DEFAULT_LANG, {}))  # fallback to English
    log.debug(f"    Triggers for node name {state_name} and lang {context['lang']}:\n{triggers}")

    buttons = node.get('buttons', {})
    buttons = buttons.get(context['lang'], buttons.get('en', {}))  # fallback to English
    log.debug(f"    Buttons for node name {state_name} and lang {context['lang']}:\n{buttons}")

    # want to check both button text values and user_intent text values
    buttons.update(user_intents)

    if user_text in buttons:
        new_state_name = buttons[user_text]
        if new_state_name:
            context['state_name'] = new_state_name
            context['bot_text'] = states[new_state_name].get('actions', {}).get(context['lang'], '')
    else:
        timeout = triggers.get('timeout')
        if isinstance(timeout, list):
            log.debug(f"Awaiting timeout {timeout}")
            await_timeout(seconds=timeout[0])
            new_state_name = timeout[1]
            context['state_name'] = new_state_name
            context['bot_text'] = states[new_state_name].get('actions', {}).get(context['lang'], '')
        elif isinstance(timeout, (float, int)):
            await_timeout(timeout)
    if list(context.keys()) == ['lang']:
        log.error(f"Intent {user_text} not found in state {node['name']}!")
    return context


def on_answer_noop(*args, **kwargs):
    log.warning('User action was a NOOP!!!!')
    log.warning(f'  NOOP\nargs: {args}\nkwargs={kwargs}')


class Engine:  # Engine(pd.Series)
    """ A state machine designed for managing the dialog for a rule-based chatbot """

    def __init__(self, states=EXAMPLE_DIALOG_TREE, **context):
        self.hist = []
        self.context = {}
        self.states = preprocess_dialog_tree_series(states)

        self.hist_append_context()  # '__init__.enter')
        self.update_context(context)

        state_name = self.context.get('state_name') or self.states.index[0]
        self.set_state_name(state_name)

        lang = self.context.get('lang') or DEFAULT_LANG
        self.set_lang(lang)

        # Unused side-effect
        actions = self.states[state_name]['actions']
        actions_with_buttons = self.states[state_name]['actions_with_buttons']
        self.bot_text_with_buttons = actions_with_buttons.get(lang, actions[lang])

    def set_state_name(self, state_name):
        """ Skip to arbitrary state_name. """
        state_name = normalize_state_name(state_name)
        if state_name is None or state_name not in self.states.index:
            log.warning(f"New state_name {state_name} not found in states.index")
            state_name = self.states.index[0]
        self.context['state_name'] = state_name  # self.default_states = {None: self.states.index[0]}
        log.info(f"Setting state_name={self.context['state_name']}")
        return state_name

    def set_lang(self, lang):
        lang = self.context.get('lang') if lang is None else DEFAULT_LANG
        lang = LANGS.get(lang)
        lang = DEFAULT_LANG if not lang else lang
        self.context['lang'] = lang
        return lang

    def update_context(self, context_changes):
        """ Update self.context with context_changes dict """
        # self.context['context_changes'] = context_changes
        self.context.update(context_changes)
        return self.context

    def normalize_user_text(self, user_text):
        return(str(user_text).lower().strip())

    def do_user_intent(self, **context):
        self.update_context(context)
        user_text = self.context.get('user_text')
        state_name = self.context.get('state_name')
        node = self.states[state_name]
        on_answer_functions = node.get('on_answer', [])
        for fun_name in on_answer_functions:
            fun = globals().get(fun_name, on_answer_noop)
            self.update_context(fun.__call__(context))
        triggers = node.get('triggers', {})
        lang = self.context.get('lang', DEFAULT_LANG)
        intents = triggers[lang]
        user_intent_text = self.normalize_user_text(user_text)
        self.set_state_name(intents[user_intent_text])
        return self.context

    def hist_append_context(self, substate=None):
        """ Append self.context to a running log of the context variable

        TODO: record only the diff/patch (`context_updates`) with `replay_context_updates()` method
        """
        if substate is not None:
            self.context['substate'] = substate
        self.hist.append(self.context)
        return self.context

    def run(self, **context_changes):
        """ Move to next state based on user action/text

        Inputs:
          user_text (str): whatever message user typed for the current state_name
          state_name (str): skip the current state and go to state_name before processing user_text
          lang (str): en, zh, zht, es, etc

        Returns dict(
          name="name-of-new-state",
          bot_text="Whatever the bot should say in response to this user action or message.",
          lang="en"|"es"|"zh"|... ,
          )
        """
        self.context = self.update_context(context_changes)  # default_lang, default_state_name
        self.hist_append_context()  # '__init__.after_update_context')
        self.update_context(context_changes)
        self.hist.append(self.context)
        self.set_state_name(self.get_next_state_name())
        return self.context

    def execute_actions(self, actions):
        if actions is None:
            return
        if 'update_context' in actions:
            self.update_context(actions['update_context'])
        if self.context['lang'] in actions:
            self.update_context(dict(bot_text=actions[self.context['lang']]))
        return self.context

    def update_intent_triggers_with_buttons(self):
        node = self.states[self.context['state_name']]
        triggers = node['triggers']
        buttons = node.get('buttons', {})
        user_intents = triggers.get(
            self.context['lang'], triggers.get(DEFAULT_LANG, {}))  # fallback to English
        log.debug(f"    Triggers for node name {self.context['state_name']} and lang {self.context['lang']}:\n{user_intents}")

        # TODO?: copy.deepcopy(buttons)
        buttons = buttons.get(self.context['lang'], buttons.get('en', {}))  # fallback to English

        # want to check both button text values and user_intent text values
        buttons.update(user_intents)
        buttons_and_intents = buttons
        log.debug(f"    Buttons and intents for node name {self.context['state_name']} and lang {self.context['lang']}:\n{buttons_and_intents}")
        return buttons_and_intents

    def detect_triggers(self):
        node = self.states[self.context['state_name']]
        triggers = node['triggers']
        state_name = self.context['state_name']
        log.debug(f"    triggers for node name {state_name}:\n{triggers}")
        context = self.context

        user_text = context.get('user_text')
        new_state_name = None

        buttons_and_intents = self.update_intent_triggers_with_buttons()

        log.warning(buttons_and_intents.keys())
        user_text = context.get('user_text', '')
        if user_text in buttons_and_intents:
            state_name = self.set_state_name(buttons_and_intents[user_text])
            node = self.states[state_name]
            self.execute_actions(node.get('actions'))
        if 'timeout' in triggers:
            if new_state_name:
                log.error('Undefined behavior when both timeout and user_text (intent) trigger a state change')
            timeout = triggers['timeout']
            if isinstance(timeout, dict):
                log.debug(f"Awaiting timeout {timeout}")
                timeout = list(list(timeout.items())[0])
            if isinstance(timeout, (list, tuple)):
                await_timeout(seconds=timeout[0])
                new_state_name = timeout[1]
                context['state_name'] = new_state_name
                context['bot_text'] = self.states[new_state_name].get('actions', {}).get(context['lang'], '')
            # FIXME: this seems useless and an anti-pattern:
            elif isinstance(timeout, (float, int)):
                await_timeout(timeout)
        if new_state_name is None:
            log.error(f"No state change was triggered by user_text {user_text} for state_name {node['name']}!")
            new_state_name = context.get('state_name')
            log.error(f"    Remaining in state_name {new_state_name}!")
        self.context = context
        return new_state_name

    def get_next_state_name(self):
        r""" Recognize desired state transition trigger and return next_state name

          Inputs:
            lang (str): e.g. 'en', 'zh', 'es'
            state name (str): e.g. 'language-selected-english'
            user utterance (str): 'Hello chatobt'
          Outputs:
            name (str): state name
            text: bot utterance

        >>> eng = Engine()

        FIXME:
        >> eng.get_next_state_name()
        'selected-language-welcome'
        """
        state_name = self.context['state_name']
        log.warning(f"Trying to find next state after {state_name}")
        # default_response = dict(
        #     state_name=states[0]['name'],
        #     bot_text=states[0].get('actions', {}).get(lang, states[0].get('en', '')),
        #     lang=lang)
        if state_name is None:
            return None
        # if str(state_name).lower().strip() in ('__start__', '__welcome__'):
        #     return default_response
        log.debug('\n' + str(self.states))

        try:
            node = self.states[state_name]
        except KeyError as e:
            log.error(e)
            log.error(f"Invalid state_name '{state_name}'. It is not among states.index:\n  {self.states.index}")

        if state_name != node['name']:
            log.debug(f' states[name]:{node.get("name")} != {state_name}:state_name')

        # detect intent from user text message
        new_state_name = self.detect_triggers()

        return new_state_name


"""
>>> from qary.chat.v3 import *
>>> e = Engine()
>>> e.run('English')
{'lang': 'en',
 'state_name': 'selected-english',
 'bot_text': 'Hello, my name is POLY - short for Polyglot, which means I can speak many languages!'
             '\n\nPlease select which language you would like me to help you in:\n'}
>>> e.state_name
'selected-english'
>>> e.states[e.state_name]
{'name': 'selected-english',
 'level': 1,
 'actions': {'update_context': {'lang': 'en'}},
 'triggers': {'timeout': {0: 'selected-language-welcome'}},
 'actions_with_buttons': {},
 'iloc': 2}
"""


default_dialog_engine = Engine()


def extract_lang(context):
    # state = context.get('state', {})
    state_name = context.get('state_name', 'unknown-state')
    context[state_name + '.answer'] = context.get('user_text', '')
    return context


def next_state(state_name='select-language', user_text='English', context=None):
    """ API example for Greg

    >>> next_state(state_name='select-language', user_text='English')
    {'state_name': 'selected-language-welcome', 'lang': 'en', 'user_text': 'English', 'select-language.answer': 'English'}
    """
    context = {} if context is None else context
    context['state_name'] = state_name
    context['user_text'] = user_text
    return default_dialog_engine.do_user_intent(**context)



""" converts yml to json"""
import json
import pathlib
import sys
import yaml

from qary import init

YAML_FILEPATH = pathlib.Path(init.DATA_DIR, 'testsets', 'dialog', 'qa-2020-04-25.yml')
JSON_FILEPATH = pathlib.Path(init.DATA_DIR, 'testsets', 'dialog', 'qa-2020-04-25.json')


def yaml_to_json(yamlpath=YAML_FILEPATH, jsonpath=JSON_FILEPATH):
    r"""converts yaml to json file

    >>> yamlpath = pathlib.Path(init.DATA_DIR, 'testsets', 'dialog', 'qa-2020-04-25.yml')
    >>> jsonpath = yamlpath.parent / (yamlpath.name[:-3] + 'json')
    >>> yaml_to_json(yamlpath=yamlpath, jsonpath=jsonpath)
    PosixPath('.../testsets/dialog/qa-2020-04-25.json')
    """

    with open(yamlpath) as yml:
        data = yaml.full_load(yml)

    with open(jsonpath, 'w') as js:
        json.dump(data, js, ensure_ascii=False, indent=2)

    return jsonpath


if __name__ == '__main__':
    yamlpath = YAML_FILEPATH
    jsonpath = JSON_FILEPATH
    if len(sys.argv) == 3:
        yamlpath = sys.argv[1]
        jsonpath = sys.argv[2]
    elif len(sys.argv) == 2:
        yamlpath = sys.argv[1]
        jsonpath = yamlpath + '.json'
    yaml_to_json(yamlpath=yamlpath, jsonpath=jsonpath)



""" String manipulation and hashing utilities for indexing Wikipedia pages and titles """
import gzip
import hashlib
from pathlib import Path
import re
import json

import numpy as np
import pandas as pd
from slugify import slugify  # noqa

from qary.config import DATA_DIR
from qary.spacy_language_model import nlp


def simple_tokenize(s: str):
    r""" Regular expression tokenizer from nlpia.ch2

    >>> ' '.join(simple_tokenize("Can't say goodbye... yet."))
    "Can't say goodbye . . . yet ."
    """
    return list(re.findall(r'\w+(?:\'\w+)?|[^\w\s]', s))


def normalize_wikititle(title: str):
    r""" Case folding and whitespace normalization for wikipedia cache titles (keys)

    >>> normalize_wikititle("\n _Hello_\t\r\n_world_!  _\n")
    'hello world !'
    """
    return re.sub(r'[\s_]+', ' ', title.title()).strip().lower()


def capitalize_acronyms(text: str):
    r""" Find acronyms and capitalize them

    >>> capitalize_acronyms("aye cap'n my ai captain")
    "aye cap'n my AI captain"
    """


def capitalize_acronyms(text: str):
    toks = nlp(text)
    toks_info = []
    for t in toks:
        d = {}
        for k in dir(t):
            d[str(k)] = str(getattr(t, k, ''))
        toks_info.append(d)
    # FIXME:
    json.dump([
        dict([
            (k, v[:70]) for (k, v) in d.items()
        ]) for d in toks_info
    ], indent=4)
    return toks_info

    # https://raw.githubusercontent.com/krishnakt031990/Crawl-Wiki-For-Acronyms/master/AcronymsFile.csv


def squash_wikititle(title: str, lowerer=str.lower):
    r""" Lowercase and remove all non-alpha characters, including whitespace

    >>> squash_wikititle("_Hello \t g00d-World_ !!!")
    'hellogdworld'
    """
    return re.sub(r'[^a-z]', '', lowerer(str(title)))


def normalize_name(s, lower=True, underscores=True, strip=True):
    r""" String normalizer to create variable name(ish) str (.lower .strip .replace \s=>_

    >>> normalize_name(' a-a . ! ')
    'a_a_._!'
    """
    if lower:
        s = str.lower(s)
    if strip:
        s = str.strip(s)
    if underscores:
        s = re.sub(r'[\s-]', '_', s)
    return s


def normalize_keys(turns_list_raw=None):
    """ Normalize the keys of a raw turns list (typically from a human-edited yaml file)

    1. Lowercase all keys in all dicts
    2. Strip whitespace from beginning and end of all keys in all dicts
    3. Replace all spaces (' ') with underscords ('_') in all keys in all dicts
    4. Rename the 'nlp' key to 'match_method'

    `turns_list_raw` must be mutable in place (list of dicts, rather than tuple of dicts)

    >>> normalize_keys(turns_list_raw=[dict(state='State Name', nlp='EXACT'), {' State  ': 'ID_01', ' NLP ': None}])
    [{'state': 'State Name', 'match_method': 'EXACT'},
     {'state': 'ID_01', 'match_method': None}]
    """
    for i, turn in enumerate(turns_list_raw):
        if not turn:  # possibility of empty list values
            continue
        turn = {(key or '').lower().strip().replace(' ', '_'): value for key, value in turn.items()}
        if 'nlp' in turn:
            turn['match_method'] = turn['nlp']
            del turn['nlp']
        turns_list_raw[i] = turn
    return turns_list_raw


def md5(s: str, num_bytes=8, dtype=np.uint64):
    r""" Like builtin hash but consistent across processes and accepts custom dtype (default npuint64)

    >>> md5('helloworld')
    6958444691603744019
    >>> md5(b'helloworld')
    6958444691603744019
    >>> md5('hello world')
    14810798070885308281
    >>> md5('Hello world')
    4747398888332172685
    >>> type(_)
    <class 'numpy.uint64'>
    >>> md5('Hello world', num_bytes=4, dtype=np.uint32)
    3659505037
    >>> md5('Hello world', num_bytes=5, dtype=np.uint32)
    3659505037
    >>> type(_)
    <class 'numpy.uint32'>
    >>> md5('Hello world', num_bytes=5, dtype=np.uint64)
    149688393101
    >>> md5('Hello world', num_bytes=4, dtype=np.uint64)
    3659505037
    """
    s = s if isinstance(s, bytes) else s.encode()
    s = str(s).encode()
    hasher = hashlib.md5()
    hasher.update(s)
    # md5.digest() is 16-byte hex representation in raw bytes object like b'^\xb6;\xbb\xe0...
    # this truncates it to the num_bytes LSBs
    hex_digest = hasher.digest()[::-1][:num_bytes]
    digest_dtype = dtype(
        np.sum(np.fromiter(
            (dtype(c) * dtype(256)**dtype(i) for i, c in enumerate(hex_digest)),
            dtype=dtype)))
    return digest_dtype


def hashed_titles_series(titles, dtype=np.uint32):
    """ Ensure titles are in np.array, normalize each title, alphabetize,

    Essentially:
        titles = titles.sort_values().unique()
        hashes = np.array([md5(t) for t in titles]).astype(dtype)
        return pd.Series(titles, index=hashes)

    >>> hashed_titles_series(['Hello', 'g00d-World!'])
    1609046494713271862       hello
    15478009694767029955    gdworld
    dtype: object
    >>> md5(squash_wikititle('!!HELLO!!')) in _.index
    True
    """
    if hasattr(titles, 'columns'):
        titles = titles[titles.columns[0]]
    if hasattr(titles, 'values'):
        titles = titles.values
    titles = np.array(titles)
    titles = np.array([squash_wikititle(str(t)) for t in titles])
    title_hashes = np.fromiter((md5(title) for title in titles), dtype=np.uint64, count=len(titles))
    return pd.Series(titles, index=title_hashes)


def read_hashes(
        filepath=Path(DATA_DIR, 'corpora', 'wikipedia', 'wikipedia-titles-alphaonly-hashed.uint64.npy.gz')):
    with gzip.open(filepath, 'rb') as fin:
        hashesread = np.load(fin)
    return hashesread



from graphviz import Digraph
from pathlib import Path

from qary.chat.dialog import TurnsPreparation, load_dialog_turns


def graphviz_dumps(graph, format='svg', quiet=True):
    """ Dump a graphviz Graph (or Digraph) object to a unicode string """
    return graph.pipe(format=format, quiet=quiet).decode()


def graph_dialog_turns(dialog_turns, name='default_dialog_graph'):
    """ Create graphviz flow chart of a list of dicts dialog.v2 dicts

    >>> graph_dialog_turns([
    ...     {'state': 'state1', 'bot': 'All good?', 'next_condition':
    ...         {'state2': 'yes', 'state3': 'N'}},
    ...     {'state': 'state2', 'bot': 'Good!'},
    ...     {'state': 'state3', 'bot': 'Sorry...'},
    ...     ])
    <graphviz...>
    """
    # if dialog_turns is None:
    #     dialog_turns = DEFAULT_DIALOG_PATH
    if isinstance(dialog_turns, (Path, str)) and Path(dialog_turns).is_file():
        dialog_turns = load_dialog_turns(dialog_turns)

    g = Digraph(name)
    g.attr(rankdir='TB')

    for turn in dialog_turns:
        source_name = turn['state']
        next_conditions = turn.get('next_condition', {'__default__': ['']})
        for dest_name, human_statements in next_conditions.items():
            if isinstance(human_statements, str):
                human_statements = [human_statements]
            label = human_statements[0]
            g.edge(source_name, dest_name, label=label)
    return g


def draw_convoscript(graph=None, dialog_turns=None, name='qary.visualizations.draw_convoscript'):
    """ Display a finite statem machine diagram of the conversation dialog graph

    >>> draw_convoscript()
    <graphviz...
    """
    if graph is None:
        if isinstance(dialog_turns, (str, Path)) and Path(dialog_turns).is_file():
            dialog_turns = TurnsPreparation(turns_list=dialog_turns)
        if dialog_turns is not None:
            g = graph_dialog_turns(dialog_turns)
    if graph is None:
        g = Digraph(name, filename=None)
    return g



""" Pattern and template based chatbot dialog engines """
import logging

from qary.chat.v3 import Engine
from qary.skills.base import normalize_replies, ContextBaseSkill


log = logging.getLogger(__name__)

dialog_enging = Engine()


class Skill(ContextBaseSkill):
    r""" Skill that can reply with answers to frequently asked questions using data/faq/*.yml

    >>> bot = Skill()
    >>> bot.reply(None)
    [(1.0, 'I\'m part of the Let\'s Talk NYC team and I\'m here to help ... a few questions? Select "Yes" when ready\n')]
    """

    def __init__(self, *args, **kwargs):
        """ Load the v3.dialog.yaml file to initialize the dialog state machine (dialog engine) """
        self.engine = Engine(*args, **kwargs)
        super().__init__(*args, **kwargs)

    def reply(self, statement, context=None):
        """ Suggest responses to a user statement string with [(score, reply_string)..]"""
        if context is None:
            context = {}
        statement = statement or ''
        # response = self.engine.run(statement, **kwargs)
        response = self.engine.run(**context)
        responses = [(1.0,
                      response.get('bot_text',
                                   'ERROR: not bot_text returned from v3.Engine()')
                      )
                     ]
        normalize_replies(responses)
        return responses



""" Pattern and template based chatbot dialog engines """
import logging
import os

print ('PATH = ' + os.environ['PATH'])

import pronouncing
import random
import re 

class Skill:
    r"""Returning a rhyming sentence"""
    def __init__(self, statement, context):
        self.statement = statement
        self.context = None

    def reply(statement):
        r"""Returns a sentence in which the final word rymes with the final word of the input string

        Examples:
            #TODO
        """
     
        # Removing any punctuation from the string using regex, except apostrophes
        statement = re.sub(r"[^\w\d'\s]+", '', statement)
        word_list = statement.split() # list of words
        rhyming_words = pronouncing.rhymes(word_list[-1]) # list of words that rhyme with the last word in 
                                                          # input sentence
        doc = nlp(statement) # tokenizing the statement for pos_ classification

        rhyme_dict = {}
        pos = doc[-1].pos_ # The part of speech of the original last word
        for rhyme in rhyming_words:
            sentence = statement.rsplit(' ', 1)[0] + ' ' + rhyme # putting rhymes in context for accurate 
                                                                 # pos_ classification
            tok = nlp(sentence) # tokenizing the modified statement for pos_ classification
            if tok[-1].pos_ == pos or tok[-1].pos_ == 'PROPN':
                rhyme_dict[rhyme] = tok[-1].pos_
        if len(rhyme_dict) == 0:
            return "Nothing rhymes with " + word_list[-1]
        choices = list(rhyme_dict.keys())
        #print(f"{pos}-Choices -----> ", choices)
        response = ' '.join(word_list[:-1]) + ' ' + str(random.choice(choices))
        return [(1.0, response)]



""" Rule-based (intent recognition and templated responses) chatbot dialog engine for quizzes """
import logging
from pathlib import Path

from qary.constants import DATA_DIR, QUIZ_DOMAINS
# from qary.skills.base import ContextBaseSkill

from qary.chat import dialog
from qary.chat.dialog import generate_domain_filepaths, get_nxt_cndn_match_mthd_dict
from qary.chat.understanders import check_match
# import TurnsPreparation, compose_statement, load_dialog_turns

# FIXME: dialog.TurnsPreparation is incompatible with the local one defined here
# from qary.chat.dialog import TurnsPreparation

# FIXME: make this a config option
# INTERN_QUIZ = os.path.join(DATA_DIR, 'yes_no/intern_quiz.yml')
INTERN_QUIZ = generate_domain_filepaths(
    domains=QUIZ_DOMAINS, prefix='', suffix='.input.dialog', ext='.yml')
INTERN_QUIZ = Path(DATA_DIR) / 'testsets/dialog_parser.input.dialog.yml'


log = logging.getLogger(__name__)


class Skill:
    r"""Skill for Quiz"""

    def __init__(self, datafile=None, turns_list=None, use_nlp=False):
        """If datafile is not given, the turns list of dicts can directly be passed to seed the data
        This would be useful for testing purposes and such
        """
        self.datafile = datafile
        self.turns = {}
        self.turns_input = turns_list
        self.use_nlp = use_nlp
        # if turns is passed, then you should not set the datafile
        if not turns_list:
            datafile = datafile or INTERN_QUIZ
            self.turns_input = dialog.load_dialog(datafile)

        if self.turns_input:
            # Do more complex operations using the helper '_TurnsPreparation' class
            turns_preparation = dialog.TurnsPreparation(turns_list=self.turns_input, use_nlp=self.use_nlp)
            self.turns = turns_preparation.prepare_turns()
        else:  # some sort of error
            log.error('An empty turns_list was passed to quiz.Skill.__init__()')
        self.state = None
        self.current_turn = None
        return

    def find_next_state(self, statement, method, nxt_cndn_match_mthd_dict):
        for next_state_option in nxt_cndn_match_mthd_dict[method]:
            match_response = check_match(
                statement, next_state_option, method
            )

            match_found = match_response[0]
            self.state = match_response[1]

            if match_found:
                return match_found

    def reply(self, statement, context=None):
        r"""Except for the welcome state, all other states are mere recordings of the quiz responses

        Examples:
            #TODO
        """
        if statement in [None, 'none']:
            statement = None

        # First check to see if we are in the time before the welcome state
        if self.state in (None, False, 0, '', ''.encode(), '0', 'none', 'None'):
            # First figure out the welcome state name using a magical special 'WELCOME' string
            # as the key. This will allow you to access the actual welcome turn
            self.state = self.turns['WELCOME']
            self.current_turn = self.turns[self.state]
            response = '\n'.join(self.current_turn['bot'])
        else:
            nxt_cndn = self.current_turn['next_condition']
            nxt_cndn_match_mthd_dict = get_nxt_cndn_match_mthd_dict(nxt_cndn)
            # for match_method_keyword in ['EXACT', '']
            match_found = False

            for matcher in 'EXACT LOWER CASE_SENSITIVE_KEYWORD KEYWORD NORMALIZE'.split():
                match_found = self.find_next_state(statement, matcher, nxt_cndn_match_mthd_dict)
                if match_found:
                    break

            if not match_found:
                self.state = nxt_cndn_match_mthd_dict[None][0][1]
            self.current_turn = self.turns.get(self.state or '', None) or None
            if self.current_turn:
                response = self.current_turn.get('bot') or ''
                if not isinstance(response, str):
                    response = '\n'.join(response)
            else:
                response = (
                    'Session is already over! Type "quit" to exit or press "Enter" for a '
                    'new session'
                )
        return [(1.0, response)]



import copy
from collections import defaultdict
import json
import logging
import math
from pathlib import Path
import yaml
import io

from ..config import DATA_DIR, DOMAINS
from ..config import DEFAULT_BOT_USERNAME
from ..init import maybe_download
from ..etl.utils import normalize_keys, normalize_name
from ..etl.nesting import default_normalizer

WELCOME_STATE_NAME = '__WELCOME__'
FINISH_STATE_NAME = EXIT_STATE_NAME = '__FINISH__'
DEFAULT_STATE_NAME = '__default__'

WELCOME_STATE_NAME = normalize_name(WELCOME_STATE_NAME)
FINISH_STATE_NAME = EXIT_STATE_NAME = normalize_name(FINISH_STATE_NAME)

INTERN_QUIZ_PATH = maybe_download(
    filename='tests/dialog_parser.v2.dialog.yml')


try:
    from collections.abc import Mapping
except ImportError:  # python <3.7
    from collections import Mapping


def generate_domain_filepaths(domains=DOMAINS, prefix=None, data_subdir='quiz', suffix='', ext='.yml'):
    quizdirpath = Path(DATA_DIR) / data_subdir
    data_subdir = data_subdir or 'quiz'
    prefix = prefix or ''
    suffix = suffix or ''
    if not ext.startswith('.'):
        ext = '.' + ext
    for domain in domains:
        pattern = f'{prefix}*{domain}*{suffix}{ext}'.replace('**', '*')
        for filepath in Path(quizdirpath).glob(pattern):
            try:
                filepointer = open(filepath)
                filepointer.close()
            except FileNotFoundError as e:
                log.error(f"{e}\n    Unable to find the file path object: {filepath}")
                break
            yield filepath


def get_nxt_cndn_match_mthd_dict(nxt_cndn):
    """Creates a dict with the match_method keyword being the key and a list of next_states which use
    that keyword as a value. This is needed because there is a priority in which the match_method
    keywords are handled

    Args:
        nxt_cndn (dict): dict with the intent being the key and the value being another dict
            with a key value pair for the next state and the 'match_method' for that intent

    Returns:
        object: A dictionary which looks like :
                { match_method1: [intent1: next_state1], } where the list is a list of all the
                next_conditions that use that match method

    """
    log.warning(f'nxt_cndn={nxt_cndn}')
    nxt_cndn_match_mthd_dict = defaultdict(list)

    for intent, nxt_state_dict in nxt_cndn.items():
        match_method = nxt_state_dict['match_method']
        next_state = nxt_state_dict['next_state']
        nxt_cndn_match_mthd_dict[match_method].append((intent, next_state))
    log.warning(f'nxt_cndn_match_mthd_dict={nxt_cndn_match_mthd_dict}')
    return nxt_cndn_match_mthd_dict


log = logging.getLogger(__name__)


####################################################################
# duplicated in src/qary/etl/nesting.py


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    Inputs:
        dct (dict): dict into which the merge is executed
        merge_dct (dict): dict merged into dct
    Returns: None (only `dct` is updated in place)

    >>> old = dict(zip('abc', '123'))
    >>> new = dict(zip('cd', '34'))
    >>> dict_merge(old, new)
    >>> old == dict(zip('abcd', '1234'))
    True
    """
    for k, v in merge_dct.items():
        # TODO: test with dest `dict` replaced with `Mapping`
        if (k in dct
            and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def dict_diff(a, b):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    Inputs:
        a (dict): original dictionary
        b (dict): changed dictionary
    Returns:
        delta_a_to_b (dict): deepcopy of b with any unchanged k, v pairs deleted

    >>> old = {'a': '1', 'b': '2', 'c': '3'}
    >>> new = {                    'c': '3', 'd': 4}
    >>> dict_diff(old, new)
    {'d': 4, 'a': None, 'b': None}
    """
    abdiff = copy.deepcopy(b)
    keys_to_del = []
    for k, v in abdiff.items():
        if k in a:
            if isinstance(b[k], dict) and isinstance(a[k], Mapping):
                # TODO: test with dest `dict` replaced with `Mapping`
                abdiff[k] = dict_diff(a[k], b[k])
            elif abdiff[k] == a[k]:
                keys_to_del.append(k)
    for k, v in a.items():
        if k not in abdiff:
            abdiff[k] = None
    for k in keys_to_del:
        del abdiff[k]
    return abdiff


normalize_key = normalize_name


def dict_replace(unclean, mapping, replace_values=True, replace_keys=False):
    """ Replace values in nested dicts according to the mapping provided

    >>> dict_replace(
    ...     {'W': {'a': 'True', 'b': True}, 'X': {'c': {False: {'d': True}}}},
    ...     mapping={False: 'n', True: 'y'})
    {'W': {'a': 'True', 'b': 'y'}, 'X': {'c': {False: {'d': 'y'}}}}
    """
    if replace_keys:
        raise NotImplementedError
    elif not replace_values:
        log.error('dict_replace() not doing anything!!!')
        return unclean
    clean = {}
    for k, v in unclean.items():
        if isinstance(unclean[k], Mapping):
            # TODO: test with dest `dict` replaced with `Mapping`
            clean[k] = dict_replace(v, mapping=mapping, replace_values=replace_values, replace_keys=replace_keys)
        else:
            clean[k] = mapping.get(v, v)
    return clean


def lod_replace(unclean, mapping, replace_values=True, replace_keys=False,
                list_types=(list, tuple), dict_types=(Mapping,)):
    """ Replace values in list of dicts according to mapping provided

    >>> lod_replace(
    ...     [{'a': 'True', 'b': True}, {'c': [False, {'d': True}]}],
    ...     mapping={False: 'n', True: 'y'})
    [{'a': 'True', 'b': 'y'}, {'c': ['n', {'d': 'y'}]}]
    """
    nonscalar_types = tuple(list(dict_types) + list(list_types))
    if not isinstance(mapping, Mapping):
        mapping = dict(mapping)
    if replace_keys:
        raise NotImplementedError
    elif not replace_values:
        log.error('dict_replace() not doing anything!!!')
        return unclean
    is_dict = isinstance(unclean, dict_types)
    is_list = isinstance(unclean, list_types)
    clean = {} if isinstance(unclean, dict_types) else []
    if is_list:
        item_generator = enumerate(unclean)
    elif is_dict:
        item_generator = unclean.items()
    else:
        return copy.copy(unclean)
    for k, v in item_generator:
        if isinstance(v, nonscalar_types):
            replaced_value = lod_replace(
                v, mapping=mapping,
                replace_values=replace_values, replace_keys=replace_keys)
        else:
            replaced_value = mapping.get(v, v)
        if is_list:
            clean.append(replaced_value)
        else:
            clean[k] = replaced_value
    return clean

# duplicated in src/qary/etl/nesting.py
####################################################################


def is_valid_turns_list(turns):
    """ Check to see if list of turns loaded from yaml file is a dialog.v2 specification

    >>> is_valid_turns_list(['whatever'])
    False
    >>> is_valid_turns_list('state_name: whatever')
    False
    >>> is_valid_turns_list([{'state': 'whatever'}])
    1
    >>> is_valid_turns_list([{'state': 'whatever'}, {}, {}])
    3
    """

    if isinstance(turns, (list, tuple)) and len(turns) > 0 and isinstance(turns[0], dict):
        return len(turns)
    return False


####################################################################
# moved from Jose's quiz.py

def load_dialog_turns(datafile):
    """Load datafile (currently yml) and create a turns datastructure

    >>> datafile = Path(DATA_DIR) / 'tests' / 'intern_quiz.yml'

    >> load_dialog_turns(datafile) # doctest: +ELLIPSIS
    [{'state': '__default__', 'bot_next_condition':...
    """
    datafile = Path(datafile)
    if not datafile.exists():
        log.error(f'Quiz bot data file {datafile} does not exist')
        turns = None
    elif datafile.suffix not in ['.yml', '.yaml']:
        log.error(f'Quiz bot currently only supports YAML datafiles with extensions yml or yaml')
        turns = None
    else:
        with open(datafile, 'r') as infile:
            turns = yaml.load(infile, Loader=yaml.SafeLoader)
    turns = lod_replace(turns, mapping={False: 'No', True: 'Yes'})
    return turns


# def normalize_keys(turns_list_raw=None):
#     """ Normalize the keys of a raw turns list (typically from a human-edited yaml file)

#     1. Lowercase all keys in all dicts
#     2. Strip whitespace from beginning and end of all keys in all dicts
#     3. Replace all spaces (' ') with underscords ('_') in all keys in all dicts
#     4. Rename the 'nlp' key to 'match_method'

#     `turns_list_raw` must be mutable in place (list of dicts, rather than tuple of dicts)

#     >>> normalize_keys(turns_list_raw=[dict(state='State Name', nlp='EXACT'), {' State  ': 'ID_01', ' NLP ': None}])
#     [{'state': 'State Name', 'match_method': 'EXACT'},
#      {'state': 'ID_01', 'match_method': None}]
#     """
#     for i, turn in enumerate(turns_list_raw):
#         if not turn:  # possibility of empty list values
#             continue
#         turn = {(key or '').lower().strip().replace(' ', '_'): value for key, value in turn.items()}
#         if 'nlp' in turn:
#             turn['match_method'] = turn['nlp']
#             del turn['nlp']
#         turns_list_raw[i] = turn
#     return turns_list_raw


def listify_bot_statements(turns_list):
    r""" Ensure that all bot statements are lists of strs (alternative bot statements)

    >>> listify_bot_statements([{'state': '1', 'bot': 'hello world'}])
    [{'state': '1', 'bot': ['hello world']}]
    >>> listify_bot_statements([{'state': '1', 'bot': [1, 2]}])
    [{'state': '1', 'bot': ['1', '2']}]
    >>> listify_bot_statements([{'state': '1', 'bot': b'hello bytes'}])
    [{'state': '1', 'bot': ['hello bytes']}]
    """
    listified_turns_list = []
    for i, turn in enumerate(turns_list):
        listified_turn = copy.deepcopy(turn)
        if DEFAULT_BOT_USERNAME in listified_turn:
            if isinstance(turn['bot'], (str, bytes)):
                bot_statements = [turn['bot']]
            else:
                bot_statements = list(turn['bot'])
            listified_turn['bot'] = []
            for statement in bot_statements:
                if isinstance(statement, bytes):
                    try:
                        listified_turn['bot'].append(statement.decode('utf-8'))
                    except UnicodeDecodeError:
                        listified_turn['bot'].append(statement.decode('latin'))
                elif isinstance(statement, str):
                    listified_turn['bot'].append(statement)
                else:
                    log.warning(f'Coercing Bot Statement from type `{type(statement)}` to `str`.')
                    listified_turn['bot'].append(str(statement))
        listified_turns_list.append(listified_turn)
    return listified_turns_list


def normalize_state_names(turns_list_raw):
    """ Normalize the state names of a raw turns list (typically from a human-edited yaml file)

    0. Run normalize_keys() on turns_list_raw to ensure the key "state" is normalized
    1. Lowercase all state names (including outgoing destination state names)
    2. Strip whitespace from beginning and end of all state names
    3. Replace all spaces (' ') with underscords ('_') in all state names
    4. Rename the 'next' key to 'match_method'

    >>> turns_list_raw=[
    ...     dict(state='State Name', nlp='EXACT', next={'ID 01': 'keyword'}),
    ...     {' State  ': 'ID_01', ' NLP ': None, 'next': ' state name '}]
    >>> normalize_keys(turns_list_raw=turns_list_raw)
    [{'state': 'State Name',
      'next': {'ID 01': 'keyword'},
      'match_method': 'EXACT'},
     {'state': 'ID_01', 'next': ' state name ', 'match_method': None}]

    """
    turns_list_raw = normalize_keys(turns_list_raw=turns_list_raw)
    for turn in turns_list_raw:
        state_orig = turn.get('state', '')
        state = (state_orig or '').lower().strip().replace(' ', '_')
        # first normalize the state name
        if state != state_orig:
            log.warning(f'Normalized state name {state_orig} to {state}')
            turn['state'] = state
        # normalize the 'next' key if it exists
        if 'next' in turn:
            turn['next'] = (turn['next'] or '').lower().strip().replace(' ', '_')
        if 'next_condition' in turn:
            next_condition_new = {}
            # create a new dictionary to avoid inplace mod of dict in loop
            for next_state, intent in turn['next_condition'].items():
                next_condition_new[(next_state or '').strip().replace(' ', '_')] = intent
            turn['next_condition'] = next_condition_new
        if 'match_method' in turn and isinstance(turn['match_method'], dict):
            match_method_new = (
                {}
            )  # create a new dictionary to avoid inplace mod of dict in loop
            for next_state, match_method in turn['match_method'].items():
                match_method_new[(next_state or '').lower().strip().replace(' ', '_')] = match_method
            turn['match_method'] = match_method_new
    return turns_list_raw


def compose_statement(statements):
    r""" Pick a sttement or combine multiple statements into one.

    Currently uses `'\n'.join(statements)`.
    Alternatively we could use the normalize_replies() method and
     probabilitistically/psychometrically chose an optimal one.

    >>> compose_statement(statements=['Hello', 'Trisolaris'])
    'Hello\nTrisolaris'
    """
    if not statements:
        return statements
    if isinstance(statements, str):
        return statements
    return '\n'.join(statements)


class TurnsPreprocessor:
    """ Prepare the turns datastructure to ensure each has a next state (outgoing graph edges)"""

    DEFAULTS = {'match_method': 'exact'}

    def __init__(self, turns_list=None, use_nlp=False, normalizer=normalize_name):
        """Raw turns that need to be cleaned up and prepared

        Args:
            use_nlp (bool)
            turns_list (list)

        Returns:
            dialog_v2_datastructure: list of dicts
        """
        self.normalizer = normalizer
        self.turns_list_input = turns_list
        self.use_nlp = use_nlp
        self.defaults = self.DEFAULTS

    def normalize_keys(self):
        """Normalize the keys of the turns dictionary as well as convert the 'nlp' key names to
        the more intuitive 'match_method'"""

        for i, turn in enumerate(self.turns_list_input):
            if not turn:  # possibility of empty list values
                continue
            # FIXME: turn = {self.normalizer(key): value for key, value in turn.items()}
            turn = {default_normalizer(key): value for key, value in turn.items()}
            if 'nlp' in turn:
                turn['match_method'] = turn['nlp']
                del turn['nlp']
            self.turns_list_input[i] = turn
        return

    def prepare_turns(self, turns_list_input=None):
        """Processes the turns to a form that allows for conditional transitions in the FSM. This
        will convert the player value in each turn from a list of dicts and strings to a pure
        dictionary, each being of the following form.
        { <intent string>: {
                'next_state': <next_state_for_that_intent>,
                'match_method': <nlp processing method for that intent>
                }
        }
        If the player repsone is empty, the key will be None."""
        if not turns_list_input:
            turns_list_input = self.turns_list_input
        self.turns_list_input = [
            turn for turn in turns_list_input if turn
        ]  # Remove empty values
        self.normalize_keys()
        self.normalize_state_names()
        self.turns_list_input = listify_bot_statements(self.turns_list_input)
        turns = {}  # convert to a dictionary with the keys being the states
        # TODO: keep track of state normalizations and if multiple states map to the same
        #  normalized state name.
        # state_normalizations = {} # keep track of state normalizations for error reporting
        self.parse_defaults()
        for i, turn_orig in enumerate(self.turns_list_input):
            # construct a new turn to avoid certain corner case errors in later processing
            turn = {}
            # Copy over any unknown keys as such for possible downstream processing
            keys_to_avoid = ['next_condition', 'next', 'state', 'match_method']
            for key, value in turn_orig.items():
                if key not in keys_to_avoid:
                    turn[key] = value
            # check if a next_condition key exists and if not create one
            next_condition = turn_orig['next_condition'] if 'next_condition' in turn_orig else {}
            next_default = self.evaluate_next_default_state(i, turn_orig)
            # now flip around the next state and the response that triggers the next state since
            # that is more useful; Also explode the multiple responses into separate keys
            next_condition_rev = self.process_next_conditions(next_condition, next_default)
            # now add nlp info for that particular intent
            next_condition_rev = self.add_match_methd_info(next_condition_rev, turn_orig)
            turn['next_condition'] = next_condition_rev
            state_name = turn_orig['state']
            turns[state_name] = turn
            if i == 0:
                # special key to indicate which is the welcome state since a dictionary does not
                # indicate which state was the first turn that was parsed
                turns[WELCOME_STATE_NAME]['state'] = state_name
            elif i == len(self.turns_list_input) - 1:
                turns[FINISH_STATE_NAME] = state_name
        self.turns_new = turns  # Add as a property for debugging ease
        return turns

    def evaluate_next_default_state(self, i, turn_orig):
        # assign the default next state to be the next sequential state if the 'next' key is
        # not present. Also unify all next_condition and next key into one dictionary
        if 'next' in turn_orig:
            state_orig = turn_orig['next']
            next_default = state_orig.lower()
            if next_default != state_orig:
                log.warning(f'Normalized "next" state name {state_orig} to {next_default}')
        else:
            if i < len(self.turns_list_input) - 1:
                state_orig = self.turns_list_input[i + 1]['state']
                next_default = state_orig.lower()
                if next_default != state_orig:
                    log.warning(f'Normalized "next" state name {state_orig} to {next_default}')
            else:
                next_default = EXIT_STATE_NAME
        return next_default

    def process_next_conditions(self, next_condition, next_default):
        next_condition_rev = {}  # keys and values reversed as well as unrolled
        for next_state, responses in next_condition.items():
            for response in responses:
                # response = normalize_wikititle(response)
                next_state_lower = next_state.lower()
                if next_state != next_state_lower:
                    log.warning(f'Normalized "next" state name {next_state} to {next_state_lower}')
                next_condition_rev[response] = {'next_state': next_state_lower}
        next_condition_rev[''] = {'next_state': next_default}
        return next_condition_rev

    def parse_defaults(self):
        """Incorporates any defaults for the yml file into the turns data structure. The first
        list item should have the defaults if any. If found these will be incorporated into the
        data structure and that particular list item will be removed"""
        defaults = self.turns_list_input[0]
        if defaults['state'] == DEFAULT_STATE_NAME:
            if ('match_method' in defaults) and (DEFAULT_STATE_NAME in defaults['match_method']):
                self.defaults['match_method'] = defaults['match_method'].get(DEFAULT_STATE_NAME).upper()
            # now remove this item from the list
            del self.turns_list_input[0]
        return

    def add_match_methd_info(self, next_condition_rev, turn_orig):
        """Adds match_method information to the turn based on a priority scheme as follows:
        - If the 'match_method' dict within the turn has a key corresponding to the next_condition state
        name, the value of that key will be used for that next_condition
        - if the next_condition state name is missing in the 'match_method' dict, then if a __default__
        key exists inside the 'match_method' dict, then that key's value will be used
        - if there is no __default__ key either in the match_method dict, then the __default__ key's value
        from that yaml file's defaults will be used, if that is also missing, then the global
        default for the match_method key (defined as a class variable)will be used. This is not explicitly
        done in this method, but the defaults property of this class would already have been
        pre-populated with the global defaults
        """

        for intent, next_state_dict in next_condition_rev.items():
            next_state = next_state_dict['next_state']
            if not intent:  # fall through state does not need a match_method
                next_state_dict['match_method'] = None
            elif 'match_method' not in turn_orig:
                next_state_dict['match_method'] = self.defaults['match_method']
            elif next_state not in turn_orig['match_method']:
                if DEFAULT_STATE_NAME in turn_orig['match_method']:
                    next_state_dict['match_method'] = turn_orig['match_method'][DEFAULT_STATE_NAME]
                else:
                    next_state_dict['match_method'] = self.defaults['match_method']
            else:
                next_state_dict['match_method'] = turn_orig['match_method'][next_state]
            # these are really constants which is made clearer by the uppercase
            if intent:
                next_state_dict['match_method'] = next_state_dict['match_method'].upper()
        return next_condition_rev

    def normalize_state_names(self):
        """Normalizes state names in various fields of the turn list by lowercasing them to keey
        things consistent"""
        for turn in self.turns_list_input:
            state_orig = turn.get('state', '')
            state = normalize_name(state_orig)
            # first normalize the state name
            if state != state_orig:
                log.warning(f'Normalized state name {state_orig} to {state}')
                turn['state'] = state
            # normalize the 'next' key if it exists
            if 'next' in turn:
                turn['next'] = turn['next'].lower()
            if 'next_condition' in turn:
                next_condition_new = {}  # create a new dictionary to avoid inplace mod of dict in loop
                for next_state, intent in turn['next_condition'].items():
                    next_condition_new[next_state.lower()] = intent
                turn['next_condition'] = next_condition_new
            if 'match_method' in turn:
                match_method_new = {}  # create a new dictionary to avoid inplace mod of dict in loop
                for next_state, match_method in turn['match_method'].items():
                    match_method_new[next_state.lower()] = match_method
                turn['match_method'] = match_method_new
        return self.turns_list_input


def load_dialog(data):
    """Load datafile (currently yml) and create a turns datastructure

    >>> intern_quiz_for_tests = Path(DATA_DIR) / 'tests' / 'intern_quiz.yml'

    >> load_dialog(intern_quiz_for_tests)  # doctest: +ELLIPSIS
    [{'state': '__default__', ...
    >> load_dialog([{'state': 'whatever'}])  # doctest: +ELLIPSIS
    [{'state': 'whatever'}]
    """
    turns = None
    if is_valid_turns_list(data):
        return data
    if isinstance(data, (str, Path)):
        filepath = Path(data)
        if not filepath.is_file():
            log.error(f'Quiz bot data file {filepath} does not exist')
        elif filepath.suffix not in ['.yml', '.yaml']:
            log.error(f'Quiz bot currently only supports YAML files with extensions .yml or .yaml')
        else:
            with open(filepath, 'r') as infile:
                return yaml.load(infile, Loader=yaml.SafeLoader)
        if not turns and isinstance(data, str):
            turns = yaml.full_load(io.StringIO(data))
            if is_valid_turns_list(turns):
                return turns
            else:
                turns = None
    return turns


class TurnsPreparation:
    """Helper class methods to prepare the turns datastructure like with next state info.

    Note: This class is merely a means to organize the various methods needed in a sane fashion
    """

    DEFAULTS = {'match_method': 'exact'}

    def __init__(self, turns_list=None, use_nlp=False):
        """Raw turns that need to be cleaned up and prepared

        Args:
            use_nlp (bool):
            turns_list (List):

        Returns:
            Dict:
        """
        self.turns_list_input = turns_list
        self.use_nlp = use_nlp
        self.defaults = self.DEFAULTS

    def normalize_keys(self):
        """Normalize the keys of the turns dictionary as well as convert the 'nlp' key names to
        the more intuitive 'match_method'"""

        for i, turn in enumerate(self.turns_list_input):
            if not turn:  # possibility of empty list values
                continue
            turn = {key.lower(): value for key, value in turn.items()}
            if 'nlp' in turn:
                turn['match_method'] = turn['nlp']
                del turn['nlp']
            self.turns_list_input[i] = turn
        return

    def prepare_turns(self):
        """Processes the turns to a form that allows for conditional transitions in the FSM. This
        will convert the player value in each turn from a list of dicts and strings to a pure
        dictionary, each being of the following form.
        { <intent string>: {
                'next_state': <next_state_for_that_intent>,
                'match_method': <nlp processing method for that intent>
                }
        }
        If the player repsone is empty, the key will be None."""

        self.turns_list_input = [
            turn for turn in self.turns_list_input if turn
        ]  # Remove empty values
        self.normalize_keys()
        self.normalize_state_names()
        for turn in self.turns_list_input:
            log.info(f'preprocessing state named {turn["state"]}')
            if turn['state'].lower().strip() == WELCOME_STATE_NAME.lower().strip():
                turn['state'] = WELCOME_STATE_NAME
                log.info(f'renaming {turn["state"]} -> {WELCOME_STATE_NAME}')
            # self.turns_list_input.append(copy.deepcopy(turn))
        turns = {}  # convert to a dictionary with the keys being the states
        # TODO: keep track of state normalizations and if multiple states map to the same
        #  normalized state name.
        # state_normalizations = {} # keep track of state normalizations for error reporting
        self.parse_defaults()
        for i, turn_orig in enumerate(self.turns_list_input):
            # construct a new turn to avoid certain corner case errors in later processing
            turn = {}
            # Copy over any unknown keys as such for possible downstream processing
            keys_to_avoid = ['next_condition', 'next', 'state', 'match_method']
            for key, value in turn_orig.items():
                if key not in keys_to_avoid:
                    turn[key] = value
            # check if a next_condition key exists and if not create one
            next_condition = turn_orig['next_condition'] if 'next_condition' in turn_orig else {}
            next_default = self.evaluate_next_default_state(i, turn_orig)
            # now flip around the next state and the response that triggers the next state since
            # that is more useful; Also explode the multiple responses into separate keys
            next_condition_rev = self.process_next_conditions(next_condition, next_default)
            # now add nlp info for that particular intent
            next_condition_rev = self.add_match_methd_info(next_condition_rev, turn_orig)
            turn['next_condition'] = next_condition_rev
            state_name = turn_orig['state']
            turns[state_name] = turn
            if i == 0:
                # special key to indicate which is the welcome state since a dictionary does not
                # indicate which state was the first turn that was parsed
                turns[WELCOME_STATE_NAME] = state_name
            elif i == len(self.turns_list_input) - 1:
                turns['FINISH'] = state_name
        self.turns_new = turns  # Add as a property for debugging ease
        return turns

    def evaluate_next_default_state(self, i, turn_orig):
        # assign the default next state to be the next sequential state if the 'next' key is
        # not present. Also unify all next_condition and next key into one dictionary
        if 'next' in turn_orig:
            state_orig = turn_orig['next']
            next_default = state_orig.lower()
            if next_default != state_orig:
                log.warning(f'Normalized "next" state name {state_orig} to {next_default}')
        else:
            if i < len(self.turns_list_input) - 1:
                state_orig = self.turns_list_input[i + 1]['state']
                next_default = state_orig.lower()
                if next_default != state_orig:
                    log.warning(f'Normalized "next" state name {state_orig} to {next_default}')
            else:
                next_default = None
        return next_default

    def process_next_conditions(self, next_condition, next_default):
        next_condition_rev = {}  # keys and values reversed as well as unrolled
        for next_state, responses in next_condition.items():
            for response in responses:
                # response = normalize_text(response)
                next_state_lower = next_state.lower()
                if next_state != next_state_lower:
                    log.warning(f'Normalized "next" state name {next_state} to {next_state_lower}')
                next_condition_rev[response] = {'next_state': next_state_lower}
        next_condition_rev[None] = {'next_state': next_default}
        return next_condition_rev

    def parse_defaults(self):
        """Incorporates any defaults for the yml file into the turns data structure. The first
        list item should have the defaults if any. If found these will be incorporated into the
        data structure and that particular list item will be removed"""
        defaults = self.turns_list_input[0]
        if defaults['state'] == '__default__':
            if ('match_method' in defaults) and ('__default__' in defaults['match_method']):
                self.defaults['match_method'] = defaults['match_method'].get('__default__').upper()
            # now remove this item from the list
            del self.turns_list_input[0]
        return

    def add_match_methd_info(self, next_condition_rev, turn_orig):
        """Adds match_method information to the turn based on a priority scheme as follows:
        - If the 'match_method' dict within the turn has a key corresponding to the next_condition state
        name, the value of that key will be used for that next_condition
        - if the next_condition state name is missing in the 'match_method' dict, then if a __default__
        key exists inside the 'match_method' dict, then that key's value will be used
        - if there is no __default__ key either in the match_method dict, then the __default__ key's value
        from that yaml file's defaults will be used, if that is also missing, then the global
        default for the match_method key (defined as a class variable)will be used. This is not explicitly
        done in this method, but the defaults property of this class would already have been
        pre-populated with the global defaults
        """

        for intent, next_state_dict in next_condition_rev.items():
            next_state = next_state_dict['next_state']
            if intent is None:  # fall through state does not need a match_method
                next_state_dict['match_method'] = None
            elif 'match_method' not in turn_orig:
                next_state_dict['match_method'] = self.defaults['match_method']
            elif next_state not in turn_orig['match_method']:
                if '__default__' in turn_orig['match_method']:
                    next_state_dict['match_method'] = turn_orig['match_method']['__default__']
                else:
                    next_state_dict['match_method'] = self.defaults['match_method']
            else:
                next_state_dict['match_method'] = turn_orig['match_method'][next_state]
            # these are really constants which is made clearer by the uppercase
            if intent:
                next_state_dict['match_method'] = next_state_dict['match_method'].upper()
        return next_condition_rev

    def normalize_state_names(self):
        """Normalizes state names in various fields of the turn list by lowercasing them to keey
        things consistent"""
        for turn in self.turns_list_input:
            state_orig = turn['state']
            state = normalize_name(state_orig)
            # first normalize the state name
            if state != state_orig:
                log.warning(f'Normalized state name {state_orig} to {state}')
                turn['state'] = state
            # normalize the 'next' key if it exists
            if 'next' in turn:
                turn['next'] = normalize_name(turn['next'])
            if 'next_condition' in turn:
                next_condition_new = (
                    {}
                )  # create a new dictionary to avoid inplace mod of dict in loop
                for next_state, intent in turn['next_condition'].items():
                    next_condition_new[normalize_name(next_state)] = intent
                turn['next_condition'] = next_condition_new
            if 'match_method' in turn:
                match_method_new = (
                    {}
                )  # create a new dictionary to avoid inplace mod of dict in loop
                for next_state, match_method in turn['match_method'].items():
                    match_method_new[next_state.lower()] = match_method
                turn['match_method'] = match_method_new
        return

# moved from Jose's skils/quiz.py
###############################################################################################


def script_to_dialog(stream, state_name_prefix='turn_id_'):
    turn_list = yaml.load(stream, Loader=yaml.SafeLoader)
    dialog_tree = [{
        # optional defaults for this yml file
        'state': '__default__',
        'nlp': {'__default__': 'exact'},
        'version': 2.0,
    }]
    print(len(turn_list))
    digits = math.ceil(math.log10(len(turn_list)))
    name_template = f'{state_name_prefix}' + '{i:0' + f'{digits}' + 'd}'
    print(name_template)
    for i, turn in enumerate(turn_list):
        name = name_template.format(i=i)
        next_name = name_template.format(i=i + 1)
        dialog_state = {}
        statements = list(turn.items())
        dialog_state['state'] = name
        dialog_state['bot'] = statements[0][1]
        dialog_state['next_condition'] = {next_name: statements[1][1]}
        dialog_tree.append(dialog_state)
    return dialog_tree


if __name__ == '__main__':
    import sys
    serializer = json.dumps
    if len(sys.argv) > 1:
        if sys.argv[1] in ('--yaml', '-y'):
            serializer = yaml.dump
        stream = open(sys.argv[1])
    else:
        stream = sys.stdin
    turn_list = script_to_dialog(stream)
    # yaml_text = stream.read()
    sys.stdout.write(serializer(turn_list))



import requests
import html2text
import logging
import xml.etree.ElementTree as ElementTree
import concurrent.futures


log = logging.getLogger(locals().get('__name__'))

FEED_URL = "https://news.google.com/rss/topics/CAAqHAgKIhZDQklTQ2pvSWJHOWpZV3hmZGpJb0FBUAE"  # local news
# world news:"https://news.google.com/rss/topics/CAAqJggKIiBDQkFTRWdvSUwyMHZNRGx1YlY4U0FtVnVHZ0pWVXlnQVAB"
TIMEOUT = 4
HEADERS = {
    "user-agent": "Mozilla/5.0 (X11; Fedora; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/91.0.4472.114 Safari/537.36",
}


def get_news_stories_text():
    stories = get_rss_feed_stories()

    with concurrent.futures.ThreadPoolExecutor(max_workers=None) as executor:
        for story in stories:
            for article_url in story["urls"]:
                executor.submit(get_article, article_url, story)

    return stories


def get_rss_feed_stories():
    response = requests.get(FEED_URL)
    tree = ElementTree.fromstring(response.text)
    channel = tree[0]

    stories = []

    for item in channel[8:]:  # skips metadata at start of feed
        title = item[0].text.rpartition("-")[0][:-1]  # removes news outlet name from headline + trailing space
        article_urls = []

        # nbsp definition needed to parse, always needs root node so added for single link stories
        xml_string = "<!DOCTYPE xml [<!ENTITY nbsp ' '>]><root>" + item[4].text + "</root>"

        # ridiculous that google doesn't escape special characters
        xml_string = xml_string.replace("&nbsp;", '')
        xml_string = xml_string.replace("&", "&amp;")

        description = ElementTree.fromstring(xml_string)[0]   # [0] index undoes root tag adding from above

        if description.tag == "ol":
            for list_item in description[:-1]:    # ordered list tag, last item is GNews story link so need to skip
                article_urls.append(list_item[0].attrib["href"])
        elif description.tag == "a":    # only one source covering the story
            article_urls.append(description.attrib["href"])
        stories.append({"title": title, "urls": article_urls, "articles": []})

    return stories


def get_article(url, story):
    text = get_article_content(url)

    if text is not None:
        story["articles"].append(text)


def get_article_content(url):
    log.info("Requesting " + url)

    try:
        response = requests.get(url, headers=HEADERS, timeout=TIMEOUT)
    except requests.exceptions.RequestException as e:
        log.debug("Timed out requesting " + url)
        log.warning(e)
        return

    if response.status_code != 200:
        log.debug("Received " + str(response.status_code) + " from " + url)
        return

    text_maker = html2text.HTML2Text()
    text_maker.ignore_links = True
    text_maker.ignore_emphasis = True
    text_maker.ignore_images = True
    text_maker.ignore_tables = True

    text = response.text
    summary = text[:256]
    return text_maker.handle(summary)



""" Transformer based chatbot dialog engine for answering questions """
import logging

import numpy as np
import transformers

from qary.config import CLI_ARGS
from qary.skills.base import BotReply
from qary.skills.base import ContextBaseSkill


log = logging.getLogger(__name__)

STOP_WIKI_PROBABILITY = .4


class Skill(ContextBaseSkill):
    """ Skill that provides answers to questions given context data containing the answer """

    def __init__(self, context=None, cli_args=CLI_ARGS, model=None, *args, **kwargs):
        """ Initialize a Question Answering skill using model supplied (default model=BERT)

        >>> qa = Skill()
        >>> callable(qa.reply)
        True
        >>> qa.model({
        ...     'question': "Who was Adam's son ?",
        ...     'context': "Cain was Abel's brother. They were both Adam's sons."})
        {'score': 0.9..., 'start': 0, 'end': 4, 'answer': 'Cain'}
        """
        context = {'doc': {'text': ''}}
        log.debug(f'Initial qa.Skill.context: {getattr(self, "context", None)}')
        self.model = transformers.pipeline('question-answering') if model is None else model
        # self.wikiscraper = wikititles.WikiScraper()
        super().__init__(context=context, cli_args=cli_args, *args, **kwargs)

    def reply(self, statement, context=None, **kwargs):
        """ Use context document + BERT to answer question in statement

        context is a nested dictionary with two ways to specify the documents for a BERT context:
        {docs: ['doc text 1', '...', ...]}
        or, as only syntactic sugar (internally this is converted to the format above)
        {doc: {text: 'document text ...'}}
        """
        statement = statement or ''
        log.info(f"QABot.reply(statement={statement}, context={context})")
        if isinstance(context, str):
            context = dict(docs=[context], doc=dict(text=context))
        elif isinstance(context, (list, tuple, np.ndarray)):
            context = dict(docs=context, doc=dict(text='\n'.join(context)))
        # this calls self.update_context(context=context) internally:
        elif isinstance(context, dict):
            if not context.get('docs') and context.get('doc'):
                context['docs'] = [context.get('doc').get('text')]

        if (not context or not any(context.get('docs', ['']))):
            gendocs = self.wikiscraper.find_article_texts(
                query=statement,
                max_articles=1, max_depth=1,
                ngrams=3,
                ignore='who what when where why'.split()) or []
            context = dict(docs=gendocs)
        responses = super().reply(
            statement=statement, context=context, **kwargs) or []
        log.info(f"qa_bots.Skill.super() responses (before BERT): {responses}")
        docs = self.context.get('docs') or [self.context['doc']['text']]
        for i, text in enumerate(docs):
            log.info(
                f"text[{i}] from context['doc']['text'] or wikipedia_api: "
                f"{repr(text)[:40]}...{repr(text)[-40:]} ({len(text)} chars)")
            super().update_context(context=dict(doc=dict(text=text)))
            if len(text.strip()) < 2:
                log.info(f'Context document text was too short: "{text}"')
                continue
            qa_inputs = {'context': text, 'question': statement}
            model_prediction_dict = self.model(qa_inputs)
            probability, response_text = model_prediction_dict['score'], model_prediction_dict['answer']

            if len(response_text) > 0:
                responses.append(
                    BotReply(
                        confidence=probability,
                        text=response_text,
                        skill=self.__module__))
                if probability > STOP_WIKI_PROBABILITY:
                    log.info(f"Short circuiting wiki crawl because p > thresh: {probability} > {STOP_WIKI_PROBABILITY}")
                    break
        return responses



import cmudict
import math
from pathlib import Path
from qary.config import DATA_DIR

import random
import pandas as pd
import re
from scipy.sparse import dok_matrix
from tqdm import tqdm  # show progress bar


df = pd.DataFrame(cmudict.entries(), columns=['word', 'phones'])
df['exact_rhyme'] = df.phones.apply(list.copy)
df['strong_rhyme'] = df.phones.apply(list.copy)
df['vowel_rhyme'] = df.phones.apply(list.copy)

vowels = ('A', 'E', 'I', 'O', 'U')


def first_consonant_removal(phones_lists):
    """ opening consonants removal

    >>> first_consonant_removal([['B', 'AW1', 'T'],['K', 'AH0', 'Z']])
    [['AW1', 'T'], ['AH0', 'Z']]
    """
    for phone_list in phones_lists:
        for i, phone in enumerate(phone_list):
            i -= i   # in order to always evaluate the first index
            if phone_list[i].startswith(vowels):
                break
            else:
                phone_list.remove(phone_list[i])

    return phones_lists


def phones_lists_to_exact_rhyme(phones_lists=df.exact_rhyme):
    """ first consonant removal, accent removal and syllable tagging

    >>> phones_lists_to_exact_rhyme([['B', 'AW1', 'T'],['K', 'AH0', 'Z']])
    [['AW_1', 'T_1'], ['AH_1', 'Z_1']]
    """
    phones_lists = first_consonant_removal(phones_lists)  # remove first consonants in place

    for phone_list in phones_lists:
        syllable = 1
        for i, phone in enumerate(reversed(phone_list)):
            if phone.startswith(vowels):
                phone_list[-(i + 1)] = phone[:-1] + '_' + str(syllable)
                syllable += 1
            else:
                phone_list[-(i + 1)] = phone + '_' + str(syllable)
    return phones_lists


def phones_lists_to_strong_rhyme(phones_lists=df.strong_rhyme):
    """ consonant removal (save for last syllable consonants), accent removal and syllable tagging

    >>> phones_lists_to_strong_rhyme([['Z', 'UW1', 'G', 'AA0', 'N', 'AA0', 'V']])
    [['UW_3', 'AA_2', 'AA_1', 'V_1']]
    """
    for phone_list in phones_lists:
        syllable = 1
        remove_counter = 0
        for i, phone in enumerate(reversed(phone_list)):
            if phone_list[-(i + 1) + remove_counter].startswith(vowels):
                phone_list[-(i + 1) + remove_counter] = phone[:-1] + '_' + str(syllable)
                syllable += 1
            elif not phone_list[-(i + 1) + remove_counter].startswith(vowels) and syllable < 2:
                phone_list[-(i + 1)] = phone + '_' + str(syllable)
            else:
                phone_list.remove(phone_list[-(i + 1) + remove_counter])
                remove_counter += 1

    return phones_lists


def phones_lists_to_vowel_rhyme(phones_lists=df.vowel_rhyme):
    """ consonant removal, accent removal and syllable tagging

    >>> phones_lists_to_vowel_rhyme([['B', 'AW1', 'T'],['K', 'AH0', 'Z']])
    [['AW_1'], ['AH_1']]
    """
    for phone_list in phones_lists:
        syllable = 1
        remove_counter = 0
        for i, phone in enumerate(reversed(phone_list)):
            if phone_list[-(i + 1) + remove_counter].startswith(vowels):
                phone_list[-(i + 1) + remove_counter] = phone[:-1] + '_' + str(syllable)
                syllable += 1
            else:
                phone_list.remove(phone_list[-(i + 1) + remove_counter])
                remove_counter += 1
    return phones_lists


df.exact_rhyme = phones_lists_to_exact_rhyme()
df.strong_rhyme = phones_lists_to_strong_rhyme()
df.vowel_rhyme = phones_lists_to_vowel_rhyme()


def word_search(source, target):
    for idx, word in enumerate([source, target]):
        if word in list(df.word) and idx == 0:
            source_vector = list(df.iloc[df.word.eq(source).idxmax(), 5:])  # series-to-vector of first source word occurrence
        elif word in list(df.word) and idx == 1:
            target_vector = list(df.iloc[df.word.eq(target).idxmax(), 5:])  # series-to-vector of first target word occurrence
        else:
            for i in range(len(word)):
                if word[:i + 1] in list(df.word) and word[i + 1:] in list(df.word):
                    left_word = df.loc[df.index[df.word == word[:i + 1]][0], 'exact_rhyme']
                    right_word = [df.loc[df.index[df.word == word[i + 1:]][0], 'phones'][0] + '_x'] + \
                        df.loc[df.index[df.word == word[i + 1:]][0], 'exact_rhyme']
                    conjunction = left_word + right_word
                    # Redoing the syllables for the conjunction
                    syllable = 1
                    for i, phone in enumerate(reversed(conjunction)):
                        if phone.startswith(vowels):
                            conjunction[-(i + 1)] = phone[:-1] + str(syllable)
                            syllable += 1
                        else:
                            conjunction[-(i + 1)] = phone[:-1] + str(syllable)
                    break
            vector = []
            for column in df.columns[5:]:
                if column in conjunction:
                    vector.append(1)
                else:
                    vector.append(0)
            if idx == 0:
                source_vector = vector
            else:
                target_vector = vector
    return source_vector, target_vector


def cosine_similarity(source, target):
    source_vector, target_vector = word_search(source, target)

    dot_prod = 0
    for i, v in enumerate(source_vector):
        dot_prod += v * target_vector[i]

        mag_1 = math.sqrt(sum([x**2 for x in source_vector]))
        mag_2 = math.sqrt(sum([x**2 for x in target_vector]))

        score = dot_prod / (mag_1 * mag_2)
    return round(score, 2)


def vowel_dictionary(phones_lists=df.vowel_rhyme):
    vowel_dict = {' '.join(k): [] for k in phones_lists}
    for idx, vows in enumerate(tqdm(phones_lists, desc="Constructing 'vowel' Dictionary", leave=False)):
        vowel_dict[' '.join(vows)].append(df.word[idx])
    return vowel_dict


def strong_dictionary(phones_lists=df.strong_rhyme):
    strong_dict = {' '.join(k): [] for k in phones_lists}
    for idx, vows in enumerate(tqdm(phones_lists, desc="Constructing 'strong' Dictionary", leave=False)):
        strong_dict[' '.join(vows)].append(df.word[idx])
    return strong_dict


def exact_dictionary(phones_lists=df.exact_rhyme):
    exact_dict = {' '.join(k): [] for k in phones_lists}
    for idx, vows in enumerate(tqdm(phones_lists, desc="Constructing 'exact' Dictionary", leave=False)):
        exact_dict[' '.join(vows)].append(df.word[idx])

    return exact_dict


def rhyme_dict(dict_type='strong'):
    if dict_type == 'exact':
        dictionary = exact_dictionary()
    elif dict_type == 'strong':
        dictionary = strong_dictionary()
    elif dict_type == 'vowel':
        dictionary = vowel_dictionary()
    else:
        raise ValueError("Invalid dictionary type; choose between 'exact', strong' or 'vowel' dict_type")
        return
    return dictionary


def rhymer(source, dict_type='strong'):
    """
    A dictionary of rhyming words. .keys() are the pnones according to dict_type (see below) and .values() are
    the words that adhere to the phones pattern

    Parameters
    ----------
    source: str, word to be rhymed with
    dict_type: str, dictionary of rhyming words to be created, default 'strong'
            'exact' => creates dict of words with all consonants, except the intial one, and vowels that rhyme with source
            'strong' => creates dict of words with all vowels and trailing consonants that rhyme with source
            'vowel' => creates dict of words with all vowels that rhyme with source
    """
    dictionary = rhyme_dict(dict_type)  # constructing the rhyme dictionary
    if type(dictionary) is not dict:
        return
    phones_type = f'{dict_type}_rhyme'
    phones = ' '.join(df.loc[df.index[df.word == source][0], phones_type])  # finding the 'dict_type' phones pattern
    candidates = dictionary[phones]  # finding the rhyming words
    candidates.remove(source)  # removing the 'source' word from the list

    return candidates


def rhyme_finder(source, scores=False, threshold=0.5):
    """
    Creates a sorted list of tuples, (rhyming_word, rhyme_score), of vowel_rhyme candidates using an exact_rhyme,
    cosine similarity threshold score

    Parameters
    ----------
    source: str, word used to generate rhyme candidates
    scores: bool, True => threshold applied with accompanied score, False => no threshold applied, default False
    threshold: float, minimum score needed to be included in the list (only applied for scores=True), default 0.5
    """
    vows = ' '.join(df[df.word == source].vowel_rhyme.values[0])
    candidates = vow_dict[vows]
    if scores:
        finalists = []
        for target in tqdm(candidates, desc="Scoring rhyme candidates", leave=False):
            score = cosine_similarity(source, target)
            if score >= threshold:
                finalists.append((target, score))
        finalists.sort(reverse=True, key=lambda x: x[1])
        finalists.remove((source, 1.0))
        candidates = finalists
    return candidates


def tokenizer(corpus):
    corpus = re.sub(r"[^\w\d'\s]+", '', corpus)  # Removing punctuation except apostrophes
    corpus = corpus.lower()
    corpus = corpus.replace('\t', ' ')
    corpus = corpus.replace('\n', ' </stop> ')  # Stop tokens
    corpus = corpus.strip().split()
    return corpus


myfile = open(Path(DATA_DIR) / 'rap' / 'rap_corpus.txt')
corpus = myfile.read()
words = tokenizer(corpus)
myfile.close()


def prior_word_matrix(words, n=1):
    word_set = list(set(words))
    word_dok = {word: i for i, word in enumerate(word_set)}
    if n > 1:
        n_grams = [' '.join(words[i:i + n]) for i, _ in enumerate(words[:-n + 1])]
    else:
        n_grams = [' '.join(words[i:i + n]) for i, _ in enumerate(words)]

    sets_count = len(list(set(n_grams)))
    prior_word_matrix = dok_matrix((sets_count, len(word_set)))

    n_grams_set = list(set(n_grams))
    n_grams_dok = {word: i for i, word in enumerate(n_grams_set)}

    for i, word in enumerate(n_grams[-1:n:-1]):
        n_grams_idx = n_grams_dok[word]
        prior_word_idx = word_dok[list(reversed(words))[i + n]]
        prior_word_matrix[n_grams_idx, prior_word_idx] += 1

    return prior_word_matrix, n_grams_dok, word_set


def word_before_n_gram(word_sequence, n=1):
    p_matrix, n_grams_dok, word_set = prior_word_matrix(words, n)
    index = n_grams_dok[word_sequence]
    prior_word_vector = p_matrix[index]
    probability = prior_word_vector / prior_word_vector.sum()  # [1, 0, 0, 2, 1] --> [0.25, 0, 0, 0.5, 0.25]

    return random.choices(word_set, probability.toarray()[0])  # selects column associated word according to
    # probability vector, eg. [0.25, 0, 0, 0.5, 0.25]


def stochastic_chain(seed, clean=True, min_length=5, max_length=10, seed_length=1):
    if min_length > max_length:
        raise ValueError("min_length cannot exceed _max_length")
        return
    current_words = seed.split(' ')
    if len(current_words) != seed_length:
        raise ValueError(f'wrong number of words, expected {seed_length}')

    sentence = seed
    dirty = ['fuck', 'shit', 'bitch', 'nigga', 'niggas', 'pussy', 'dick']

    if not clean:
        for _ in range(max_length - len(current_words)):
            next_word = word_before_n_gram(' '.join(current_words), seed_length)
            if next_word[0] == '</stop>' and len(sentence.split()) >= min_length:
                break
            else:
                sentence = next_word[0] + ' ' + sentence
                current_words = current_words[::-1]  # reverse the word order
                current_words = next_word + current_words[1:]
    else:
        for _ in range(max_length - len(current_words)):
            next_word = word_before_n_gram(' '.join(current_words), seed_length)
            if next_word in dirty:
                print('DIRTY!')
                continue
            elif next_word[0] == '</stop>' and len(sentence.split()) >= min_length:
                break
            else:
                sentence = next_word[0] + ' ' + sentence
                current_words = current_words[::-1]  # reverse the word order
                current_words = next_word + current_words[1:]
    if len(sentence.split()) > min_length:  # prevents n-grams from being truncated
        return sentence.replace('</stop>', '')
    else:
        return sentence



""" Breadth first exploration (crawling) of a graph of web pages (Wikipedia Articles) """
import logging
import re
import time
from collections import abc

from tqdm import tqdm
import wikipedia as wiki

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

PROSOCIAL_PAGE_TITLES = [
    "Prosocial behavior",
    "Rogerian_argument",
    "Civil discourse",
    "Epistemic_humility",
    "Epistemic_virtue",
    "Occam's razor",
    "Wisdom",
    "Understanding",
    "Discernment",
    "Altruism",
    "Omnibenevolence",
    "Social market economy"
]


class WikipediaPage(wiki.WikipediaPage):

    def __init__(self):
        """ Augment the WikipediaPage with a see_also attribute listing lines from that section """
        super().__init__()
        see_also = []
        append_see_also = False
        for line in self.content.splitlines():
            if line.startswith('=='):
                if re.match(r'[=]+\s*see also\s*[=]+', line.lower().strip()):
                    append_see_also = True
                elif append_see_also:
                    break
                continue
            elif append_see_also:
                see_also.append(line)
        self.see_also = see_also


def get_page(title):
    try:
        return wiki.page(title, auto_suggest=False)
    except (wiki.DisambiguationError, wiki.PageError) as e:
        log.warning(f'auto_suggest=False: {e}')
    try:
        return wiki.page(title, auto_suggest=True)
    except (wiki.DisambiguationError, wiki.PageError) as e:
        log.warning(f'auto_suggest=True: {e}')
    return False


def walk_wikipedia(pages, depth=1, delay=0.1):
    depth_goal = depth
    depth = 0
    if isinstance(pages, str):
        pages = {t: None for t in pages.split(',')}
    if not isinstance(pages, abc.Mapping):
        pages = {t: None for t in pages.split(',')}
    queue = list(pages.keys())
    while depth < depth_goal and len(queue):
        log.info(f"depth={depth}, len(nextqueue)={len(queue)}, queue[0]={queue[0]}")
        pages_this_level = len(queue)
        for i in tqdm(range(pages_this_level)):
            title = queue.pop(0)
            page_tuple = pages.get(title)
            # if title from front of queue has not already been retrieved, then retrieve it
            if page_tuple is None:  # if False then retrieval has been attempted and failed
                time.sleep(delay)
                page = get_page(title)
                log.debug(page)
                # FIXME: add depth_id (depth) and link_id (i) integers as attributes to page object (no tuple for values)
                pages[title] = (depth, i, page)
                if page:  # page is False when get_page() failed to find a valid page
                    queue.extend(list(page.links))
            elif page_tuple and page_tuple[-1]:
                queue.extend(list(page_tuple[-1].links))
        depth += 1
    return pages
